#ifndef ARRAY_H
#define ARRAY_H

#include <algorithm>
#include "assert.h"

//*****************************************************************************
// Class Template for a Generic Resizable N Dimensional Array        (for N>=2)
// By Giovanni Bavestrelli                  Copyright 1999 Giovanni Bavestrelli 
// Any feedback is welcome, you can contact me at gbavestrelli@yahoo.com
//
// This is the full implementation of my array, up to the new C++ standard.
// It uses partial specialization, and it will not work with Visual C++.
// If you use Visual C++, take the other version of the classes.
//
// This version has the following modification:
// I derived class Array<T,N> from class RefArray<T,N>, as this has the
//    following advantages:
// 1) I can pass an Array<T,N> to any function that expects a RefArray<T,N>
// 2) I can reuse the implementation of many members of RefArray<T,N> without
//    having to repeat them in class Array<T,N>
// The price I pay is that class Array contains two extra pointers (8 bytes)
// Note that I cannot define a (virtual) destructor in class RefArray for
// performance reasons, so this solution is not ideal.
//*****************************************************************************

// Forward declaration needed for friend declarations
template <typename T, unsigned int N>
class Array;

//=============================================================================
// Classes for passing a typesafe vector of dimensions to the Array constructor
//=============================================================================

// Class that encapsulates a const unsigned int (&)[N]
template <unsigned int N>
class ArraySize
{
     typedef const unsigned int (&UIntArrayN)[N];

     unsigned int m_Dimensions[N];

     ArraySize(const unsigned int (&Dimensions)[N-1],unsigned int dim)
     {
         std::copy(&Dimensions[0],&Dimensions[N-1],m_Dimensions);
         m_Dimensions[N-1]=dim;
     }

  public:

     ArraySize<N+1> operator () (unsigned int dim) 
     { 
        return ArraySize<N+1>(m_Dimensions,dim);
     }

     operator UIntArrayN () const { return m_Dimensions; }

     friend class ArraySizes;
     friend class ArraySize<N-1>;
};

// Starting point to build a const unsigned int (&)[N] on the fly
class ArraySizes
{
     unsigned int m_Dimensions[1];   

  public:
      
     explicit ArraySizes(unsigned int dim) 
     { 
        m_Dimensions[0]=dim;
     }

     ArraySize<2> operator () (unsigned int dim) 
     { 
        return ArraySize<2>(m_Dimensions,dim);
     }
};

//-----------------------------------------------------------------------------
// Class Template for N Dimensional SubArrays within an Array
//-----------------------------------------------------------------------------

template <typename T, unsigned int N>
class RefArray
{
 public:

   // STL-like types
   typedef T         value_type;
   typedef T &       reference;
   typedef const T & const_reference;
   typedef T *       pointer;
   typedef const T * const_pointer;
   typedef T *       iterator;
   typedef const T * const_iterator;
   typedef size_t    size_type;
   typedef ptrdiff_t difference_type;

   // Give access to number of dimensions
   enum  { array_dims = N };

 protected:

   const size_type * const m_pNDimensions; // Array dimensions
   const size_type * const m_pSubArrayLen; // SubArray dimensions

   T * m_pArrayElements; // Point to C array with elements within Array

   RefArray<T,N>(T * pElements, const size_type * pNDimensions, 
                                const size_type * pSubArrayLen)
     :m_pArrayElements(pElements),m_pNDimensions(pNDimensions),
                                  m_pSubArrayLen(pSubArrayLen)
   { 
      assert(m_pNDimensions && m_pSubArrayLen); 
      assert(!m_pArrayElements || m_pNDimensions[0]>0 && m_pSubArrayLen[0]>0); 
   }  

   // I cannot add a (virtual) destructor for performance reasons

 public:

   RefArray<T,N-1> operator [](size_type Index)
   {
      assert(m_pArrayElements);
      assert(Index<m_pNDimensions[0]);
      return RefArray<T,N-1>(&m_pArrayElements[Index*m_pSubArrayLen[0]],
                              m_pNDimensions+1,m_pSubArrayLen+1);
   }

   const RefArray<T,N-1> operator [](size_type Index) const
   {
      assert(m_pArrayElements);
      assert(Index<m_pNDimensions[0]);
      return RefArray<T,N-1>(&m_pArrayElements[Index*m_pSubArrayLen[0]],
                              m_pNDimensions+1,m_pSubArrayLen+1);
   }

   // Return STL-like iterators
   iterator       begin()       { return m_pArrayElements; }
   const_iterator begin() const { return m_pArrayElements; }
   iterator       end()         { return m_pArrayElements+size(); }
   const_iterator end()   const { return m_pArrayElements+size(); } 

   // Return size of array
   size_type size() const { return m_pNDimensions[0]*m_pSubArrayLen[0]; }

   // Return size of subdimensions
   size_type size(unsigned int Dim) const 
   {  
      assert(Dim>=1 && Dim<=N); 
      return m_pNDimensions[Dim-1];  
   }

   // Return number of dimensions 
   unsigned int dimensions()  const { return N; }

   // Say if the array is empty
   bool empty() const { return m_pArrayElements==NULL; }

 protected:

   // The following are protected mainly because they are not exception-safe
   // but the way they are used in the rest of the class is exception-safe

   // Copy the elements of another subarray on this one where possible
   // Where not possible, initialize them to a specified value Init
   void copy(const RefArray<T,N> & SA, const T & Init=T())
   {
       size_type below=std::min(size(1),SA.size(1));
       size_type above=size(1);

       // Copy the elements we can copy
       for (size_type i=0;i<below;i++) 
           (*this)[i].copy(SA[i],Init);

       // Reset the elements we can't copy
       for (size_type j=below;j<above;j++)
           (*this)[j].initialize(Init);
   }

   // Reset all the elements
   void initialize(const T & Init=T())
   {
       std::fill(begin(),end(),Init);
   }

   // Prefer non-member operator ==, but it needs to be a friend 
   template <typename TT, unsigned int NN> 
   friend bool operator == (const RefArray<TT,NN> & A, const RefArray<TT,NN> & B);

   friend class Array<T,N+1>; 
   friend class RefArray<T,N+1>; 
};


//-----------------------------------------------------------------------------
// Partial Specialization for Monodimensional SubArray within an Array
//-----------------------------------------------------------------------------

template <typename T>
class RefArray<T,1>
{
 public:

   // STL-like types
   typedef T         value_type;
   typedef T &       reference;
   typedef const T & const_reference;
   typedef T *       pointer;
   typedef const T * const_pointer;
   typedef T *       iterator;
   typedef const T * const_iterator;
   typedef size_t    size_type;
   typedef ptrdiff_t difference_type;

   // Give access to number of dimensions
   enum  { array_dims = 1 };

 protected:

   const size_type * const m_pNDimensions; // Array dimension

   T * m_pArrayElements; // Point to elements within Array

   RefArray<T,1>(T * pElements, const size_type * pNDimensions, 
                                const size_type * pSubArrayLen)
     :m_pArrayElements(pElements),m_pNDimensions(pNDimensions)
   { 
      assert(m_pNDimensions && pSubArrayLen); 
      assert(!m_pArrayElements || m_pNDimensions[0]>0 && pSubArrayLen[0]==1); // We found the elements
   } 

   // I cannot add a (virtual) destructor for performance reasons

 public:

   reference operator [](size_type Index)
   {
      assert(m_pArrayElements);
      assert(Index<m_pNDimensions[0]);
      return m_pArrayElements[Index];
   }

   const_reference operator [](size_type Index) const
   {
      assert(m_pArrayElements);
      assert(Index<m_pNDimensions[0]);
      return m_pArrayElements[Index];
   }

   // Return STL-like iterators
   iterator       begin()       { return m_pArrayElements; }
   const_iterator begin() const { return m_pArrayElements; }
   iterator       end()         { return m_pArrayElements+size(); }
   const_iterator end()   const { return m_pArrayElements+size(); }

   // Return size of array
   size_type size()  const { return m_pNDimensions[0]; } 

   // Return size of subdimensions
   size_type size(unsigned int Dim) const 
   {  
      assert(Dim==1);
      return m_pNDimensions[0];  
   }

   // Return number of dimensions 
   unsigned int dimensions()  const { return 1; }

   // Say if the array is empty
   bool empty() const { return m_pArrayElements==NULL; }

 protected:

   // The following are protected mainly because they are not exception-safe
   // but the way they are used in the rest of the class is exception-safe

   // Copy the elements of another subarray on this one where possible
   // Where not possible, initialize them to a specified value Init
   void copy(const RefArray<T,1> & SA, const T & Init=T())
   {
       size_type below=std::min(size(1),SA.size(1));
       size_type above=size(1);

       // Copy the elements we can copy
       for (size_type i=0;i<below;i++)
          m_pArrayElements[i]=SA.m_pArrayElements[i];

       // Reset the elements we can't copy
       for (size_type j=below;j<above;j++)
          m_pArrayElements[j]=Init;
   }

      // Reset all the elements
   void initialize(const T & Init=T())
   {
       std::fill(begin(),end(),Init);
   }

   // Prefer non-member operator ==, but it needs to be a friend 
   template <typename TT, unsigned int NN> 
   friend bool operator == (const RefArray<TT,NN> & A, const RefArray<TT,NN> & B);

   friend class Array<T,2>;
   friend class RefArray<T,2>; 
};


//-----------------------------------------------------------------------------
// Class Template for a Generic Resizable N Dimensional Array
//-----------------------------------------------------------------------------

template <typename T, unsigned int N>
class Array : public RefArray<T,N>
{
   // No need to repeat all typedefs etc. from RefArray<T,N>
   // except for size_type, otherwise Comeau compiler complains
   typedef size_t    size_type;

   size_type  m_NDimensions[N];  // Size of the N array dimensions
   size_type  m_SubArrayLen[N];  // Size of each subarray

 public:

   // Default constructor
   Array<T,N>()
      :RefArray<T,N>(NULL,m_NDimensions,m_SubArrayLen)  
   {
      std::fill(m_NDimensions,m_NDimensions+N,0);
      std::fill(m_SubArrayLen,m_SubArrayLen+N,0);
   }
   
   // This takes an array of N values representing the size of the N dimensions
   // You must remove the explicit keyword if you use MetroWerks CodeWarrior
   explicit Array<T,N>(const unsigned int (&Dimensions)[N], const T & Init=T())
      :RefArray<T,N>(NULL,m_NDimensions,m_SubArrayLen)  
   { 
      std::fill(m_NDimensions,m_NDimensions+N,0);
      std::fill(m_SubArrayLen,m_SubArrayLen+N,0);

      resize(Dimensions,Init);
   }

   // Copy constructor
   Array<T,N>(const Array<T,N> & A)
      :RefArray<T,N>(NULL,m_NDimensions,m_SubArrayLen)  
   {
      std::fill(m_NDimensions,m_NDimensions+N,0);
      std::fill(m_SubArrayLen,m_SubArrayLen+N,0);

      Array<T,N> Temp;
      if (!A.empty() && Temp.resize(A.m_NDimensions))
         std::copy(A.begin(),A.end(),Temp.begin());
      swap(Temp);
   }

   // Destructor
  ~Array<T,N>() 
   { 
      delete [] m_pArrayElements;
   }

   // Set the size of each array dimension
   bool resize(const unsigned int (&Dimensions)[N], const T & Init=T(), bool PreserveElems=false)
   {
      Array<T,N> Temp; 

      // Calculate all the information you need to use the array
      size_type nArrayElements=1;
      for (int i=0;i<N;i++)
      {
         if (Dimensions[i]==0)
            return false; // Check that no dimension was zero 
         nArrayElements*=Dimensions[i]; 
         Temp.m_NDimensions[i]=Dimensions[i];
         Temp.m_SubArrayLen[i]=1;              
         for (int k=N-1;k>i;k--)
            Temp.m_SubArrayLen[i]*=Dimensions[k];
      }  

      // Allocate new elements, let exception propagate
      Temp.m_pArrayElements=new T[nArrayElements];

      // Some compilers might not throw exception if allocation fails
      if (!Temp.m_pArrayElements)
         return false;

      // Copy the elements from the previous array if requested
      if (PreserveElems && !empty())
         Temp.copy(*this,Init);
      // Otherwise initialize them to the specified value
      else 
         Temp.initialize(Init);

      // Now swap this object with the temporary
      swap(Temp);

      return true; 
   }

   // Delete the complete Array
   void clear()
   { 
      delete [] m_pArrayElements; 
      m_pArrayElements=NULL; 

      std::fill(m_NDimensions,m_NDimensions+N,0);
      std::fill(m_SubArrayLen,m_SubArrayLen+N,0);
   }

   // Assignment operator
   Array<T,N> & operator = (const Array<T,N> & A)
   {
      if (&A!=this) // For efficiency  
      {
        Array<T,N> Temp(A);
        swap(Temp);
      }
      return *this;
   }

   // Swap this array with another, a'la STL 
   void swap(Array<T,N> & A)
   {
      std::swap(m_pArrayElements,A.m_pArrayElements); 
      std::swap_ranges(m_NDimensions,m_NDimensions+N,A.m_NDimensions);
      std::swap_ranges(m_SubArrayLen,m_SubArrayLen+N,A.m_SubArrayLen);
   }
};


// Test for equality between two arrays
template <typename T, unsigned int N>
bool operator == (const RefArray<T,N> & A, const RefArray<T,N> & B)
{
   return std::equal(A.m_pNDimensions,A.m_pNDimensions+N,B.m_pNDimensions)
       && std::equal(A.begin(),A.end(),B.begin());
}

// Test for inequality between two arrays
template <typename T, unsigned int N>
bool operator != (const RefArray<T,N> & A, const RefArray<T,N> & B) 
{
   return !(A==B); 
}


// Not implemented, meaningless to have 0 dimensions
template <typename T>
class Array<T,0>
{
};


// Not implemented, use std::vector for one dimensional arrays
template <typename T>
class Array<T,1>
{
};

#endif

